# -*- Mode: Python -*-

import os
import nodes
import lisp_reader
import graph
import analyze
import cps
import transform

is_a = isinstance
from pdb import set_trace as trace
from pprint import pprint as pp

# what's the idea?
#
# well, we want to prototype a byte-code compiler that will eventually
#  be written in Irken.  While doing this, we will write the VM for this
#  bytecode, which we can test using this module.
#
# now, what language shall we compile?
#
# We *could* start leaning toward python right now, but a much easier
#  path would be to use the Irken front end.  How about this:
#
# reader => transform => [skip typer] => analyze => cps
#
# Since we plan to implement an untyped language, we skip the typing phase.
#
# [we might also want to skip analyze/inlining?  can we?]

class context:
    # maintain some context between passes
    def __init__ (self):
        self.datatypes = {}

def compile_file (f, name, verbose=True):
    base, ext = os.path.splitext (name)
    print 'read...'
    r = lisp_reader.reader (f)
    exp = r.read_all()

    if verbose:
        print '--- read ---'
        pp (exp)

    c = context()

    print 'transform...'
    t = transform.transformer (c)
    exp2 = t.go (exp)
    if verbose:
        print '--- transform ---'
        pp (exp2)

    w = nodes.walker (c)
    exp3 = w.go (exp2)

    print 'rename...'
    # alpha conversion
    c.var_dict = nodes.rename_variables (exp3, c)
    # find strongly connected components
    print 'call graph...'
    c.dep_graph = graph.build_dependency_graph (exp3)
    c.scc_graph, c.scc_map = graph.strongly (c.dep_graph)

    a = analyze.analyzer (c, True, verbose)
    exp4 = a.analyze (exp3)

    if verbose:
        print '--- analyzer ---'
        exp4.pprint()

    ic = byte_cps (c, verbose=verbose)
    exp5 = ic.go (exp4)

    if verbose:
        print '--- cps ---'
        cps.pretty_print (exp5)

    fo = open ('%s.byc' % base, 'wb')
    num_regs = cps.the_register_allocator.max_reg
    b = compiler (fo, name, num_regs, c)
    b.emit (exp5)
    b.done()
    fo.close()

class byte_cps (cps.cps):
    # here, we'll override Irken behavior to get bytecode behavior

    def compile_primapp (self, tail_pos, exp, lenv, k):
        return self.compile_primargs (exp.args, (exp.name,), lenv, k)

class compiler:
    # byte-code compiler for vm/vm.scm (irken)
    def __init__ (self, fo, name, nregs, c):
        self.fo = fo
        self.name = name
        self.nregs = nregs
        self.context = c

    def write (self, s):
        self.fo.write (s)

    def done (self):
        self.fo.close()

    def emit (self, insns):
        for insn in insns:
            name = 'insn_%s' % (insn.name,)
            fun = getattr (self, name, None)
            fun (insn)

    def encode_lit (self, n):
        if n < 0:
            raise NotImplementedError
        else:
            bytes = []
            while n:
                bytes.insert (0, chr (n & 0xff))
                n >>= 8
            return bytes

    def insn_lit (self, insn):
        lit = insn.params
        insn.target
        bytes = self.encode_lit (lit)
        # ['L' <target> <llen> <l0> <l1> ...]
        self.write ('L%c%c%s' % (insn.target, len(bytes), ''.join (bytes)))

    def insn_return (self, insn):
        val_reg = insn.regs[0]
        self.write ('R%c' % (val_reg))
    
    def insn_primop (self, insn):
        prim, = insn.params
        regs = insn.regs
        if prim == '%+':
            self.write ('+%c%c%c' % (insn.target, regs[0], regs[1]))
        elif prim == '%-':
            self.write ('-%c%c%c' % (insn.target, regs[0], regs[1]))
        elif prim == '%=':
            self.write ('=%c%c%c' % (insn.target, regs[0], regs[1]))
        else:
            raise ValueError ("unknown primop %r" % (prim,))

    def insn_test (self, insn):
        ignore, then_code, else_code = insn.params
        self.write ('?%c' % (insn.regs[0],))
        self.emit (then_code)
        self.write ('.')
        self.emit (else_code)
        self.write ('.')

if __name__ == '__main__':
    import sys
    name = sys.argv[1]
    f = open (name, 'rb')
    compile_file (f, name)
